/*
 *    TTTTTTTTTTTTTT  HH          HH  OOOOOOOOOOOOOO  MM          MM
 *    TTTTTTTTTTTTTT  HH          HH  OOOOOOOOOOOOOO  MMM        MMM
 *          TT        HH          HH  OO          OO  MMMM      MMMM 
 *          TT        HH          HH  OO          OO  MM MM    MM MM
 *          TT        HH          HH  OO          OO  MM  MM  MM  MM 
 *          TT        HHHHHHHHHHHHHH  OO          OO  MM   MMMM   MM
 *          TT        HHHHHHHHHHHHHH  OO          OO  MM    MM    MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OOOOOOOOOOOOOO  MM          MM
 *          TT        HH          HH  OOOOOOOOOOOOOO  MM          MM 
 *
 *                      l'mulateur Thomson TO7-70
 *
 *  Copyright (C) 1996 Sylvain Huet, 1999-2002 Eric Botcazou.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 *  Module     : alleg/mouse.c
 *  Version    : 1.5.5
 *  Cr par   : Eric Botcazou 1999
 *  Modifi par: Eric Botcazou 03/05/2002
 *
 *  Interface du crayon optique.
 */


#ifndef SCAN_DEPEND
   #include <allegro.h>
#endif

#include "alleg/graphic.h"
#include "to7.h"


#define SCR40_SIZE   8
#define SCR80_SIZE  16

static int screen_ratio;
static int border_supported;
static BITMAP *pen_pointer;
static char scr40_pen_pointer_data[]={ 0,0,1,1,1,0,0,0,
                                       0,0,1,2,1,0,0,0,
                                       1,1,1,2,1,1,1,0,
                                       1,2,2,2,2,2,1,0,
                                       1,1,1,2,1,1,1,0,
                                       0,0,1,2,1,0,0,0,
                                       0,0,1,1,1,0,0,0,
                                       0,0,0,0,0,0,0,0 };

static char scr80_pen_pointer_data[]={ 0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,0,
                                       1,2,2,2,2,2,2,2,2,2,2,2,2,2,1,0,
                                       1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,
                                       0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };


/* Helper pour cliper les coordonnes:
 */
static inline int clip_x(void)
{
    if (border_supported)
    {
        int x = (mouse_x/screen_ratio) - TO7_BORDER_W;

        if (x<0)
            return 0;
        else if (x>=TO7_WINDOW_W)
            return TO7_WINDOW_W;
        else 
            return x;
    }
    else
        return mouse_x/screen_ratio;
}


static inline int clip_y(void)
{
    if (border_supported)
    {
        int y = (mouse_y/screen_ratio) - TO7_BORDER_W;

        if (y<0)
            return 0;
        else if (y>=TO7_WINDOW_H)
            return TO7_WINDOW_H;
        else 
            return y;
    }
    else
        return mouse_y/screen_ratio;
}


/* LightpenCallback:
 *  Routine de gestion du crayon optique du TO7-70 appele par
 *  l'interruption matrielle souris INT 0x33 via Allegro.
 */
static void LightpenCallback(int flags)
{
    static int pointer_on=1, xpos, ypos;

    if (flags&MOUSE_FLAG_LEFT_UP)
        to7_HandleMouseClick(1, TRUE);
    if (flags&MOUSE_FLAG_LEFT_DOWN)
        to7_HandleMouseClick(1, FALSE);

    if (flags&(MOUSE_FLAG_MOVE|MOUSE_FLAG_RIGHT_DOWN))
    {
        if (pointer_on)
            RetraceScreen( xpos-pen_pointer->w/2+1, ypos-pen_pointer->h/2+1,
                           pen_pointer->w         , pen_pointer->h );
                           
        to7_HandleMouseMotion(clip_x(), clip_y());

        if (flags&MOUSE_FLAG_RIGHT_DOWN)
            pointer_on^=1;

        if (pointer_on)
        {
            xpos = mouse_x;
            ypos = mouse_y;
            draw_sprite(screen, pen_pointer, xpos-pen_pointer->w/2+1,
                                             ypos-pen_pointer->h/2+1 );
        }
    }
}

END_OF_FUNCTION(LightpenCallback)



/* InstallPointer:
 *  Installe le crayon optique du TO7-70.
 */
void InstallPointer(void)
{
    mouse_callback=LightpenCallback;
}



/* ShutDownPointer:
 *  Dsactive le crayon optique du TO7-70.
 */
void ShutDownPointer(void)
{
    mouse_callback=NULL;
}



/* InitPointer:
 *  Initialise le module crayon optique.
 */
void InitPointer(int gfx_mode, int bsupp)
{
    register int i;

    switch (gfx_mode)
    {
        case GFX_WINDOW:
        case GFX_MODE80:
            /* cration du pointeur crayon optique */
            pen_pointer=create_system_bitmap(SCR80_SIZE, SCR80_SIZE);

            for (i=0; i<SCR80_SIZE*SCR80_SIZE; i++)
                switch (scr80_pen_pointer_data[i])
                {
                    case 1:
                        putpixel(pen_pointer, i%SCR80_SIZE, i/SCR80_SIZE, 1);
                        break;

                    case 2:
                        putpixel(pen_pointer, i%SCR80_SIZE, i/SCR80_SIZE, 2);
                        break;

                    default:
                        putpixel(pen_pointer, i%SCR80_SIZE, i/SCR80_SIZE, 0);
                }

            screen_ratio = 2;
            break;

        case GFX_MODE40:
            /* cration du pointeur crayon optique */
            pen_pointer=create_system_bitmap(SCR40_SIZE, SCR40_SIZE);

            for (i=0; i<SCR40_SIZE*SCR40_SIZE; i++)
                switch (scr40_pen_pointer_data[i])
                {
                    case 1:
                        putpixel(pen_pointer, i%SCR40_SIZE, i/SCR40_SIZE, 1);
                        break;

                    case 2:
                        putpixel(pen_pointer, i%SCR40_SIZE, i/SCR40_SIZE, 2);
                        break;

                    default:
                        putpixel(pen_pointer, i%SCR40_SIZE, i/SCR40_SIZE, 0);
                }
            
            screen_ratio = 1;
            break;
    }

    border_supported = bsupp;

    install_mouse();

    lock_bitmap(pen_pointer);
    LOCK_FUNCTION(LightpenCallback);
}
