/*
 *    TTTTTTTTTTTTTT  HH          HH  OOOOOOOOOOOOOO  MM          MM
 *    TTTTTTTTTTTTTT  HH          HH  OOOOOOOOOOOOOO  MMM        MMM
 *          TT        HH          HH  OO          OO  MMMM      MMMM 
 *          TT        HH          HH  OO          OO  MM MM    MM MM
 *          TT        HH          HH  OO          OO  MM  MM  MM  MM 
 *          TT        HHHHHHHHHHHHHH  OO          OO  MM   MMMM   MM
 *          TT        HHHHHHHHHHHHHH  OO          OO  MM    MM    MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OOOOOOOOOOOOOO  MM          MM
 *          TT        HH          HH  OOOOOOOOOOOOOO  MM          MM 
 *
 *                      l'mulateur Thomson TO7-70
 *
 *  Copyright (C) 1996 Sylvain Huet, 1999-2002 Eric Botcazou.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 *  Module     : dos/main.c
 *  Version    : 1.5.5
 *  Cr par   : Eric Botcazou juin 1999
 *  Modifi par: Eric Botcazou 03/05/2002
 *
 *  Boucle principale de l'mulateur.
 */


#ifndef SCAN_DEPEND
   #include <stdio.h>
   #include <stdlib.h>
   #include <string.h>
   #include <allegro.h>
#endif

#include "alleg/graphic.h"
#include "alleg/gui.h"
#include "alleg/keybint.h"
#include "alleg/main.h"
#include "alleg/mouse.h"
#include "alleg/sound.h"
#include "dos/debug.h"
#include "dos/disk.h"
#include "to7.h"


/* pour limiter la taille de l'xcutable */
BEGIN_COLOR_DEPTH_LIST
    COLOR_DEPTH_8
END_COLOR_DEPTH_LIST

BEGIN_GFX_DRIVER_LIST
    GFX_DRIVER_VGA
END_GFX_DRIVER_LIST

BEGIN_DIGI_DRIVER_LIST
    DIGI_DRIVER_SOUNDSCAPE
    DIGI_DRIVER_AUDIODRIVE
    DIGI_DRIVER_WINSOUNDSYS
    DIGI_DRIVER_SB
END_DIGI_DRIVER_LIST 

BEGIN_MIDI_DRIVER_LIST
END_MIDI_DRIVER_LIST

BEGIN_JOYSTICK_DRIVER_LIST
END_JOYSTICK_DRIVER_LIST


struct EmuTO thom={
    TRUE,
    TRUE,
    NONE
};

static int frame;           /* compteur de frame vido */
static volatile int tick;   /* compteur du timer */


static void Timer(void)
{
    tick++;
}

static END_OF_FUNCTION(Timer)



/* RunTO7:
 *  Boucle principale de l'mulateur.
 */
static void RunTO7(void)
{
    if (!thom.sound_enabled)
	install_int_ex(Timer,BPS_TO_TIMER(TO7_FRAME_FREQ));

    frame=1;
    to7_ColdReset();

    do  /* boucle principale de l'mulateur */
    {
        thom.command=NONE;
	tick=frame;

        /* installation des handlers clavier, souris et son */
        InstallKeybint();
	InstallPointer();

	if (thom.sound_enabled && thom.exact_speed)
            StartSound();
    	
        do  /* boucle d'mulation */
        {
            to7_DoFrame();

            /* rafrachissement du pourtour */
            if (need_border_refresh)
                RefreshBorder();

            /* synchronisation sur frquence relle */
            if (thom.exact_speed)
            {
                if (thom.sound_enabled) 
                    PlaySoundBuffer();
                else
                    while (frame==tick)
	   	        ;
            }

            frame++;
        }
	while (thom.command==NONE);  /* fin de la boucle d'mulation */

        /* dsinstallation des handlers clavier, souris et son */
        if (thom.sound_enabled && thom.exact_speed)
            StopSound();
	
	ShutDownPointer();
	ShutDownKeybint();

	if (thom.command==CONTROL_PANEL)
	    ControlPanel();

        if (thom.command==SCREENSHOT)
	    Screenshot();    

	if (thom.command==DEBUGGER)
	{
	     remove_keyboard();
	     SetGraphicMode(SHUTDOWN);
	     Debugger();
	     SetGraphicMode(RESTORE);
	     install_keyboard();
	}

	if (thom.command==RESET)
	    to7_Reset();

	if (thom.command==COLD_RESET)
	    to7_ColdReset();
    }
    while (thom.command != QUIT);  /* fin de la boucle principale */

    if (!thom.sound_enabled)
        remove_int(Timer);
}



/* ExitMessage:
 *  Affiche un message de sortie et sort du programme.
 */
static void ExitMessage(const char msg[])
{
    fprintf(stderr, "%s\n", msg);
    exit(EXIT_FAILURE);
}


#define IS_5_INCHES(drive) ((drive_type[drive]>0) && (drive_type[drive]<3))

/* main:
 *  Point d'entre du programme appel par MS-DOS.
 */
int main(int argc, char *argv[])
{
    char version_name[] = "Thom "TO7_VERSION_STR" (MSDOS/DPMI)";
    char memo_name[FILENAME_LENGTH] = "\0";
    int disk_ctrl = TO7_CD90_640;
    int direct_support = 0, direct_write_support = FALSE;
    int drive_type[4];
    int i;

    for (i=1;i<argc;i++)
    {
        if (!strcmp(argv[i],"-help") || !strcmp(argv[i],"-h"))
        {
            printf("Usage:\n");
            printf("           %s [options ...]\n",argv[0]);
            printf("o les options sont prises parmi les suivantes:\n");
            printf("    -help                   affiche cette aide\n");
            printf("    -m file.m7              place la cartouche file.m7 dans le lecteur\n");
            printf("    -nodisk                 dsactive le contrleur de disquettes\n");
            printf("    -fast                   active la pleine vitesse de l'mulateur\n");
            printf("    -nosound                supprime le son de l'mulateur\n");
            exit(EXIT_SUCCESS);
        }
	else if (!strcmp(argv[i],"-m"))
	    strcpy(memo_name, argv[++i]);
        else if (!strcmp(argv[i],"-nodisk"))
            disk_ctrl = TO7_NO_DISK_CTRL;
        else if (!strcmp(argv[i],"-fast"))
            thom.exact_speed = FALSE;
        else if (!strcmp(argv[i],"-nosound"))
            thom.sound_enabled = FALSE;
        else if (!strcmp(argv[i],"--enable-direct-write"))
            direct_write_support = TRUE;
        else
        {
            fprintf(stderr, "%s: [-help] [-m file.m7] [-nodisk] [-fast] [-nosound]\n",argv[0]);
            exit(EXIT_FAILURE);
        }
    }

    /* initialisation de la librairie Allegro */
    set_uformat(U_ASCII);  /* pour les accents franais */
    allegro_init();
    install_keyboard();
    install_timer();
    LOCK_VARIABLE(thom);
    LOCK_VARIABLE(tick);
    LOCK_FUNCTION(Timer);

    printf("Voici %s l'mulateur Thomson TO7-70.\n", version_name);
    printf("Copyright (C) 1996 Sylvain Huet, 1999-2003 Eric Botcazou.\n\n");
    printf("Touches: [ESC] Panneau de contrle\n");
    printf("         [F11] Capture d'cran\n");
    printf("         [F12] Dbogueur\n\n");

    /* initialisation de l'mulateur */
    printf("Initialisation de l'mulateur... ");

    if (to7_Init(disk_ctrl) == TO7_ERROR)
    {
        /* la ROM contrleur disk est peut-tre manquante? */
        disk_ctrl = TO7_NO_DISK_CTRL;

        if (to7_Init(disk_ctrl) == TO7_ERROR)
            ExitMessage(to7_error_msg);
        else
        {
            printf("%s\n", to7_error_msg);
            printf("Initialisation de l'mulateur sans support des disquettes... ");
        }
    }

    printf("ok\n");

    /* chargement de la cartouche */
    if (memo_name[0])
    {
        to7_LoadMemo7(memo_name);
    }
    else if (disk_ctrl == TO7_NO_DISK_CTRL)
    {
        to7_LoadMemo7("./memo7/basic.m7");
    }
    else
    {
        if (to7_LoadMemo7("./memo7/basic128.m7") == TO7_ERROR)
            to7_LoadMemo7("./memo7/basic.m7");
    }

    /* initialisation de l'interface clavier */
    InitKeybint();

    /* initialisation de l'interface d'accs direct */
    if (disk_ctrl != TO7_NO_DISK_CTRL)
    {
        InitDirectDisk(drive_type, direct_write_support);

        /* Dtection des lecteurs supports par le CD90-640 (5"25 seulement) */
        for (i=0; i<4; i++)
        {
            if (IS_5_INCHES(i))
                direct_support |= (1<<i);
        }
    }

    /* initialisation du son */
    InitSound(25600);  /* 25600 Hz */

    printf("\nAppuyez sur une touche");
    readkey();

    /* initialisation du mode graphique */
    InitGraphic(GFX_MODE40);

    /* initialisation de l'interface utilisateur */
    InitGUI(version_name, disk_ctrl != TO7_NO_DISK_CTRL, direct_support);

    /* initialisation de la souris */
    InitPointer(GFX_MODE40, FALSE);

    /* et c'est parti !!! */
    RunTO7();

    /* sortie du mode graphique */
    SetGraphicMode(SHUTDOWN);

    /* mise au repos de l'interface d'accs direct */
    ExitDirectDisk();

    /* sortie de l'mulateur */
    printf("A bientt !\n");
    return 0;
}

