/*
 *    TTTTTTTTTTTTTT  HH          HH  OOOOOOOOOOOOOO  MM          MM
 *    TTTTTTTTTTTTTT  HH          HH  OOOOOOOOOOOOOO  MMM        MMM
 *          TT        HH          HH  OO          OO  MMMM      MMMM 
 *          TT        HH          HH  OO          OO  MM MM    MM MM
 *          TT        HH          HH  OO          OO  MM  MM  MM  MM 
 *          TT        HHHHHHHHHHHHHH  OO          OO  MM   MMMM   MM
 *          TT        HHHHHHHHHHHHHH  OO          OO  MM    MM    MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OOOOOOOOOOOOOO  MM          MM
 *          TT        HH          HH  OOOOOOOOOOOOOO  MM          MM 
 *
 *                      l'mulateur Thomson TO7-70
 *
 *  Copyright (C) 1996 Sylvain Huet, 1999-2002 Eric Botcazou.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 *  Module     : win/graphic.c
 *  Version    : 1.5.5
 *  Cr par   : Gilles Ftis
 *  Modifi par: Eric Botcazou 15/01/2002
 *
 *  Gestion de l'affichage du TO7-70.
 */


#ifndef SCAN_DEPEND
   #include <allegro.h>
#endif

#include "alleg/graphic.h"
#include "alleg/main.h"
#include "alleg/mouse.h"
#include "to7.h"


struct SCREEN_PARAMS
{
    int allegro_drv;
    int size;
    int screen_w;
    int screen_h;
    int border_w;
    int border_h;
};

static const struct SCREEN_PARAMS scr_mode40 = { GFX_AUTODETECT_FULLSCREEN,
                                                 1,
                                                 TO7_WINDOW_W,
                                                 TO7_WINDOW_H,
                                                 0,
                                                 0 },

                                  scr_mode80 = { GFX_AUTODETECT_FULLSCREEN,
                                                 2,
                                                 TO7_WINDOW_W*2,
                                                 TO7_WINDOW_H*2,
                                                 0,
                                                 0},

                                  scr_window = { GFX_AUTODETECT_WINDOWED,
                                                 2,
                                                 TO7_SCREEN_W*2,
                                                 TO7_SCREEN_H*2,
                                                 TO7_BORDER_W*2,
                                                 TO7_BORDER_H*2 };


/* variables globales */
int need_border_refresh;

#define COLOR_SHIFT  17
static PALETTE palette={ { 0, 0, 0},  /* bordure de l'cran */
                         { 0, 0, 0},  /*     */
                         {60,63,63},  /*     */
                         { 0, 0, 0},  /* GUI */
                         {47,50,50},  /*     */
                         {60,63,63},  /*     */
                         { 0,63, 0}   /* LED verte */
};

static int graphic_mode;
static const struct SCREEN_PARAMS *scr;
static int border_color;
static BITMAP *gpl_buffer, *screen_buffer;
static int *dirty_cell;



/* RefreshBorder:
 *  Change la couleur du pourtour de l'cran.
 */
void RefreshBorder(void)
{
    RetraceScreen(scr->border_w, 0, scr->screen_w-scr->border_w, scr->border_w);
    RetraceScreen(scr->border_w, scr->screen_h-scr->border_h, scr->screen_w-scr->border_w, scr->screen_h);
    RetraceScreen(0, 0, scr->border_w, scr->screen_h);
    RetraceScreen(scr->screen_w-scr->border_w, 0, scr->screen_w, scr->screen_h);

    need_border_refresh = FALSE;
}


#define RECTF(x1, y1, x2, y2)  rectfill(screen_buffer, (x1), (y1), (x2), (y2), border_color)

/* SetBorderColor:
 *  Fixe la couleur du pourtour de l'cran.
 */
static void SetBorderColor(int color)
{
    color += COLOR_SHIFT;

    if ((scr->border_w != 0) && (border_color != color))
    {
        border_color=color;
        RECTF(scr->border_w, 0, scr->screen_w-scr->border_w-1, scr->border_w-1);
        RECTF(scr->border_w, scr->screen_h-scr->border_h, scr->screen_w-scr->border_w-1, scr->screen_h-1);
        RECTF(0, 0, scr->border_w-1, scr->screen_h-1);
        RECTF(scr->screen_w-scr->border_w, 0, scr->screen_w-1, scr->screen_h-1);
        need_border_refresh = TRUE;
    }
}



/* DrawGPL:
 *  Affiche un Groupe Point Ligne (un octet de VRAM).
 */
static void DrawGPL(int addr, int pt, int col)
{
    register int i;
    unsigned int c1, c2, x, y;
             int *dirty_cell_row;
             int gpl_size;
    unsigned char *gpl_src = gpl_buffer->line[0], *gpl_dest;

    c1=((col>>3)&7)+(((~col)&0x40)>>3)+COLOR_SHIFT;
    c2=(col&7)+(((~col)&0x80)>>4)+COLOR_SHIFT;

    if (scr->size == 1)
    {
        for (i=0; i<8; i++)
           *gpl_src++ = (0x80>>i)&pt ? c1 : c2;

        gpl_size = TO7_GPL_SIZE;
    }
    else
    {
        for (i=0; i<8; i++)
        {
           *gpl_src++ = (0x80>>i)&pt ? c1 : c2;
           *gpl_src++ = (0x80>>i)&pt ? c1 : c2;
        }

        gpl_size = TO7_GPL_SIZE*2;
    }

    x = scr->border_w + (addr%TO7_WINDOW_GW)*gpl_size;
    y = scr->border_h + (addr/TO7_WINDOW_GW)*scr->size;

    gpl_src  = gpl_buffer->line[0];
    gpl_dest = screen_buffer->line[y]+x;

    if (memcmp(gpl_dest, gpl_src, gpl_size))
    {
        memcpy(gpl_dest, gpl_src, gpl_size);

        if (scr->size == 2)
        {
            gpl_dest = screen_buffer->line[y+1]+x;
            memcpy(gpl_dest, gpl_src, gpl_size);
        }

        /* dirty rectangles */
        x = addr%TO7_WINDOW_CW;
        y = addr/(TO7_WINDOW_CW*TO7_CHAR_SIZE);
        dirty_cell_row = dirty_cell + y*TO7_WINDOW_CW;
        dirty_cell_row[x] = TRUE;
    }
}



/* RetraceScreen:
 *  Rafrachit une portion de l'cran du TO8.
 */
void RetraceScreen(int x, int y, int width, int height)
{
    blit(screen_buffer, screen, x, y, x, y, width, height);
}



/* RefreshScreen:
 *  Rafrachit l'cran du TO7-70.
 */
void RefreshScreen(void)
{
    register int i,j;
             int cell_start;
             int *dirty_cell_row = dirty_cell;
             int cell_size = (scr->size == 1 ? TO7_CHAR_SIZE : TO7_CHAR_SIZE*2);

    if (!graphic_mode)
        return;

    acquire_screen();

    /* on groupe les dirty rectangles ligne par ligne */ 
    for (j=0; j<TO7_WINDOW_CH; j++)
    {
        for (i=0; i<TO7_WINDOW_CW; i++)
            if (dirty_cell_row[i])
            {
                cell_start=i;

                while ((i<TO7_WINDOW_CW) && dirty_cell_row[i])
                    dirty_cell_row[i++]=FALSE;

                RetraceScreen(scr->border_w + cell_start*cell_size, scr->border_h + j*cell_size,
                              (i-cell_start)*cell_size, cell_size);
            }

        /* ligne suivante */
        dirty_cell_row += TO7_WINDOW_CW;
    }

    release_screen();
}



/* SetGraphicMode:
 *  Slectionne le mode graphique de l'mulateur.
 */
int SetGraphicMode(int mode)
{
    switch (mode)
    {
        case INIT:
            if (set_gfx_mode(scr->allegro_drv, scr->screen_w, scr->screen_h, 0, 0))
                return FALSE;

            set_palette(palette);
            graphic_mode=TRUE;
            break;

        case RESTORE:
            set_gfx_mode(scr->allegro_drv, scr->screen_w, scr->screen_h, 0, 0);
            set_palette(palette);
            blit(screen_buffer, screen, 0, 0, 0, 0, scr->screen_w, scr->screen_h);
            graphic_mode=TRUE;
            break;

        case SHUTDOWN:
            get_palette(palette); /* on sauvegarde la palette */
            set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
            graphic_mode=FALSE;
            break;

    } /* end of switch */

    return TRUE;
}


#define LED_SIZE 6


/* SetDiskLed:
 *  Allume/teint la LED du lecteur de disquettes.
 */
static void SetDiskLed(int led_on)
{
    int led_size = (scr->size == 1 ? LED_SIZE : LED_SIZE*2);

    if (graphic_mode)
    {
        if (led_on)
        {
            rect    (screen, scr->screen_w-led_size  , 0, scr->screen_w-1, led_size-1, 1);
            rectfill(screen, scr->screen_w-led_size+1, 1, scr->screen_w-2, led_size-2, 6);
        }
        else
            RetraceScreen(scr->screen_w-led_size, 0, led_size, led_size);
    }
}



/* InitGraphic:
 *  Initialise le pilote graphique.
 */
int InitGraphic(int gfx_mode)
{
    int i;

    set_color_depth(8);

    switch (gfx_mode)
    {
        case GFX_MODE40:
            scr = &scr_mode40;
            break;

        case GFX_MODE80:
            scr = &scr_mode80;
            break;

        case GFX_WINDOW:
            scr = &scr_window;
            break;

        default:
            return FALSE;
    }

    /* rcupration de la palette du TO7-70 */
    for (i=0; i<TO7_NCOLORS; i++)
    {
        int red, green, blue;

        to7_QueryColor(i, &red, &green, &blue);

        palette[i+COLOR_SHIFT].r = red >> 2;
        palette[i+COLOR_SHIFT].g = green >> 2;
        palette[i+COLOR_SHIFT].b = blue >> 2;
    }

    /* pointeur de la souris de la GUI */
    palette[16] = palette[1];
    palette[255] = palette[2];

    if (!SetGraphicMode(INIT))
        return FALSE;

    gpl_buffer = create_bitmap(TO7_GPL_SIZE * scr->size, 1);
    screen_buffer = create_bitmap(scr->screen_w, scr->screen_h);
    clear_bitmap(screen_buffer);
    dirty_cell = calloc(TO7_WINDOW_CW*TO7_WINDOW_CH, sizeof(int));

    to7_DrawGPL = DrawGPL;
    to7_SetBorderColor = SetBorderColor;
    to7_SetDiskLed = SetDiskLed;

    /* initialisation de la souris */
    InitPointer(gfx_mode, gfx_mode == GFX_WINDOW);

    return TRUE;
}



/* Screenshot:
 *  Prend une photo d'cran de l'mulateur.
 */
void Screenshot(void)
{
    static int count;
    char filename[FILENAME_LENGTH];
    BITMAP *bmp;
    PALETTE palette;

    get_palette(palette);
    bmp = create_sub_bitmap(screen, 0, 0, SCREEN_W, SCREEN_H);
    sprintf(filename, "thom_%02d.bmp", count++);
    save_bmp(filename, bmp, palette);
    destroy_bitmap(bmp);
}

